#ifndef VECTOR3_H_
#define VECTOR3_H_

/*! \struct Vector3
\brief Trojrozmrn (3D) vektor.

Implementace tslokovho relnho vektoru podporujc zkladn
matematick operace.

\note
Vektor se povauje za sloupcov, pestoe je v komentch pro jednoduchost
uvdn jako dkov.

\code{.cpp}
Vector3 v = Vector3( 2.0f, 4.5f, 7.8f );
v.Normalize();
\endcode

\author Tom Fabin
\version 0.9
\date 2007-2012
*/
struct /*ALIGN*/ Vector3
{
public:
	union	// anonymn unie
	{
		struct
		{
			float x; /*!< Prvn sloka vektoru. */
			float y; /*!< Druh sloka vektoru. */
			float z; /*!< Tet sloka vektoru. */
		};

		float data[3]; /*!< Pole sloek vektoru. */
	};

	//! Vchoz konstruktor.
	/*!
	Inicializuje vechny sloky vektoru na hodnotu nula,
	\f$\mathbf{v}=\mathbf{0}\f$.
	*/
	Vector3() : x( 0 ), y( 0 ), z( 0 ) { }	

	//! Obecn konstruktor.
	/*!
	Inicializuje sloky vektoru podle zadanch hodnot parametr,
	\f$\mathbf{v}=(x,y,z)\f$.

	\param x prvn sloka vektoru.
	\param y druh sloka vektoru.
	\param z tet sloka vektoru.
	*/
	Vector3( const float x, const float y, const float z ) : x( x ), y( y ), z( z ) { }

	//! L2-norma vektoru.
	/*!
	\return x Hodnotu \f$\mathbf{||v||}=\sqrt{x^2+y^2+z^2}\f$.
	*/
	float L2Norm() const;

	//! Druh mocnina L2-normy vektoru.
	/*!
	\return Hodnotu \f$\mathbf{||v||^2}=x^2+y^2+z^2\f$.
	*/
	float SqrL2Norm() const;

	//! Normalizace vektoru.
	/*!
	Po proveden operace bude mt vektor jednotkovou dlku.
	*/
	void Normalize();

	//! Vektorov souin.
	/*!
	\param v vektor \f$\mathbf{v}\f$.

	\return Vektor \f$(\mathbf{u}_x \mathbf{v}_z - \mathbf{u}_z \mathbf{v}_y,
	\mathbf{u}_z \mathbf{v}_x - \mathbf{u}_x \mathbf{v}_z,
	\mathbf{u}_x \mathbf{v}_y - \mathbf{u}_y \mathbf{v}_x)\f$.
	*/
	Vector3 CrossProduct( const Vector3 & v );	

	//! Skalrn souin.
	/*!		
	\return Hodnotu \f$\mathbf{u}_x \mathbf{v}_x + \mathbf{u}_y \mathbf{v}_y + \mathbf{u}_z \mathbf{v}_z)\f$.
	*/
	float DotProduct( const Vector3 & v ) const;

	//! Rotace.
	/*!		
	\return Vektor orotovan kolem potku o \f$\phi\f$ radin v kladnm smyslu.
	*/
	//Vector2 Rotate( const float phi );

	//! Index nejvt sloky vektoru.
	/*!
	\param absolute_value index bude uren podle absolutn hodnoty sloky

	\return Index nejvt sloky vektoru.
	*/
	char LargestComponent( const bool absolute_value = false );	

	//! Index nejmen sloky vektoru.
	/*!
	\param absolute_value index bude uren podle absolutn hodnoty sloky

	\return Index nejmen sloky vektoru.
	*/
	char SmallestComponent( const bool absolute_value = false );

	//! Element-wise absolute value.
	/*!	
	\return Element-wise absolute value.
	*/
	Vector3 Abs() const;

	void Print();

	// --- opertory ------

	friend Vector3 operator-( const Vector3 & v );

	friend Vector3 operator+( const Vector3 & u, const Vector3 & v );
	friend Vector3 operator-( const Vector3 & u, const Vector3 & v );

	friend Vector3 operator*( const Vector3 & v, const float a );	
	friend Vector3 operator*( const float a, const Vector3 & v );
	friend Vector3 operator*( const Vector3 & u, const Vector3 & v );

	friend Vector3 operator/( const Vector3 & v, const float a );
	friend Vector3 operator/( const Vector3 & u, const Vector3 & v );
	friend Vector3 operator/( const float a, const Vector3 & u );

	friend void operator+=( Vector3 & u, const Vector3 & v );
	friend void operator-=( Vector3 & u, const Vector3 & v );
	friend void operator*=( Vector3 & v, const float a );
	friend void operator/=( Vector3 & v, const float a );	
};

#endif
